// Copyright (2009) Bojan Nikolic <b.nikolic@mrao.cam.ac.uk>
// 
// This file is part of libAIR/casaWVR
//
// libAIR/casaWVR is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
// 
// libAIR/casaWVR is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with libAIR/casaWVR.  If not, see
// <http://www.gnu.org/licenses/>.
//
// Comments/bug reports/other correspondance welcome at: Bojan Nikolic
// <b.nikolic@mrao.cam.ac.uk>

#include <set>

#include <boost/scoped_ptr.hpp>
#include <boost/program_options.hpp>
#include <boost/format.hpp>
#include <boost/assign/list_inserter.hpp> // for 'push_back()'
#include <boost/assign/list_of.hpp>       // for 'list_of()' and 'ref_list_of()'
#include <boost/assert.hpp>
#include <boost/algorithm/string.hpp>

#include <hdf5.h>
#include <hdf5_hl.h>

#include <ms/MeasurementSets/MeasurementSet.h>
#include <ms/MeasurementSets/MSColumns.h>
#include <ms/MeasurementSets/MSMainEnums.h>
#include <tables/Tables/Table.h>

#include "src/apps/arraydata.hpp"
#include "casawvr/mswvrdata.hpp"
#include "casawvr/msspec.hpp"
#include "casawvr/msutils.hpp"
#include "casawvr/msantdata.hpp"
#include "../hd5wvr/hdfutils.hpp"

void listSPWs(const casa::MeasurementSet &ms,
	      std::ostream &os)
{
  os<<"--------------------- SPW listing"
    <<std::endl;

  casa::MSSpectralWindow specTable(ms.spectralWindow());

  casa::ROArrayColumn<casa::Double> inchanfreq(specTable,
					       casa::MSSpectralWindow::columnName(casa::MSSpectralWindow::CHAN_FREQ));

  casa::ROScalarColumn<casa::Double> intotbw
    (specTable,
     casa::MSSpectralWindow::columnName(casa::MSSpectralWindow::TOTAL_BANDWIDTH));

  casa::ROArrayColumn<casa::Double> chw
    (specTable,
     casa::MSSpectralWindow::columnName(casa::MSSpectralWindow::CHAN_WIDTH));
  
  for(size_t i=0; i<specTable.nrow(); ++i)
  {
    //os<<inchanfreq(i)<<", "<<intotbw(i)
    os<<chw(i)<<", "<<intotbw(i)
      <<std::endl;
  }
}

/** List the SPWs associated with WVRs
 */
void listWVRSPWs(const casa::MeasurementSet &ms,
		 std::ostream &os)
{
  using namespace LibAIR;

  SPWSet wvrs=WVRSPWIDs(ms);
  
  os<<"--------------------- SPW associated with WVRs ------------"
    <<std::endl;
  
  for(SPWSet::const_iterator i=wvrs.begin(); 
      i != wvrs.end();
      ++i)
  {
    os<<"SPW "<<*i
      <<std::endl;
  }
  

}

size_t nSPWs(const casa::MeasurementSet &ms)
{
  return ms.spectralWindow().nrow();
}

size_t nChannels(const casa::MeasurementSet &ms,
		 size_t spw)
{
  casa::ROScalarColumn<casa::Int> nc
    (ms.spectralWindow(),
     casa::MSSpectralWindow::columnName(casa::MSSpectralWindow::NUM_CHAN));

  return nc(spw);
}

void mkSPWList(const casa::MeasurementSet &ms,
	       const boost::program_options::variables_map &vm,
	       std::vector<size_t> &spwl)
{
  if (vm.count("spw"))
  {
    spwl.push_back(vm["spw"].as<size_t>());
  }
  else
  {
    for(size_t cspw=0; cspw < nSPWs(ms); ++cspw)
      spwl.push_back(cspw);
  }
}


/// Dump information obout the frequency setup of the observations
void dumpFQ(const casa::MeasurementSet &ms,
	    hid_t fout,
	    const boost::program_options::variables_map &vm)
{
  
  casa::MSSpectralWindow specTable(ms.spectralWindow());
  
  casa::ROArrayColumn<casa::Double> 
    inchanfreq(specTable,
	       casa::MSSpectralWindow::columnName(casa::MSSpectralWindow::CHAN_FREQ));
  
  casa::ROArrayColumn<casa::Double> chw
    (specTable,
     casa::MSSpectralWindow::columnName(casa::MSSpectralWindow::CHAN_WIDTH));

  std::vector<size_t> spwl;
  mkSPWList(ms, vm, spwl);


  for(std::vector<size_t>::const_iterator si=spwl.begin(); 
      si != spwl.end() ; 
      ++si)
  {
    const size_t cspw = *si;

    const size_t nch=nChannels(ms, cspw);

    std::string group;
    if (vm.count("spw"))
    {
      group=((boost::format("/ALMACORR/SPW%i/CONF") % 1).str());
    }
    else
    {
      group=((boost::format("/ALMACORR/SPW%i/CONF") % cspw).str());
    }
    H5Gcreate(fout,
	      group.c_str(),
	      NULL, NULL, NULL);

    std::vector<double> v_chfreq(nch);
    std::vector<double> v_chbw(nch);
          
    casa::Array<casa::Double> sw_freq, sw_w;
    inchanfreq.get(cspw, sw_freq, casa::True);
    chw.get(cspw, sw_w, casa::True);

    for (size_t k=0; k<nch; ++k)
    {
      v_chfreq[k]=sw_freq(casa::IPosition(1,k));
      v_chbw[k]=sw_w(casa::IPosition(1,k));
    }

    hd5::write(fout,
	       group+"/CH_FREQ",
	       v_chfreq);
  }
  
}
void infoMain(const casa::MeasurementSet &ms,
	      std::ostream &os)
{
  casa::ROArrayColumn<casa::Complex> indata(ms, 
					    casa::MS::columnName(casa::MS::DATA));
  const size_t nrows=indata.nrow();

  os<<"Number of rows: "<<nrows
    <<std::endl;

}

void dumpMain(const casa::MeasurementSet &ms,
	      hid_t fout,
	      const boost::program_options::variables_map &vm)
{

  casa::MSMainEnums::PredefinedColumns ocol;
  if (vm.count("corrected"))
  {
    ocol=casa::MS::CORRECTED_DATA;
  }
  else
  {
    ocol=casa::MS::DATA;
  }

  casa::ROArrayColumn<casa::Complex> indata(ms, 
					    casa::MS::columnName(ocol));

  /// This is the data description column
  casa::ROScalarColumn<casa::Int> indd(ms,
				       casa::MS::columnName(casa::MS::DATA_DESC_ID));

  casa::ROScalarColumn<casa::Int> a1(ms,
				     casa::MS::columnName(casa::MS::ANTENNA1));
  casa::ROScalarColumn<casa::Int> a2(ms,
				     casa::MS::columnName(casa::MS::ANTENNA2));

  casa::ROScalarColumn<casa::Int> state(ms,
					casa::MS::columnName(casa::MS::STATE_ID));

  const casa::ROMSMainColumns cols(ms);
  const casa::ROScalarColumn<casa::Int> &field=cols.fieldId();

  casa::ROScalarColumn<casa::Double> time(ms,
					  casa::MS::columnName(casa::MS::TIME));

  const size_t nrows=indata.nrow();

  H5Gcreate(fout,
	    "/ALMACORR",
	    NULL, NULL, NULL);

  std::vector<int> v_a1(nrows);
  std::vector<int> v_a2(nrows);
  std::vector<int> v_field(nrows);
  std::vector<double> v_time(nrows);
  std::vector<int> v_state(nrows);

  std::vector<size_t> spwl;
  mkSPWList(ms, vm, spwl);

  // This is the map between spectral windows and data description IDs
  std::map<size_t, size_t> ddmap=LibAIR::SPWDataDescMap(ms);


  for(std::vector<size_t>::const_iterator si=spwl.begin(); 
      si != spwl.end() ; 
      ++si)
  {
    const size_t cspw = *si;
    size_t nch=nChannels(ms, cspw);
    const size_t cdd=ddmap[cspw];

    std::string group;
    if (vm.count("spw"))
    {
      group=((boost::format("/ALMACORR/SPW%i") % 1).str());
    }
    else
    {
      group=((boost::format("/ALMACORR/SPW%i") % cspw).str());
    }
    H5Gcreate(fout,
	      group.c_str(),
	      NULL, NULL, NULL);


    std::vector<double> data(2*nrows*nch);

    const size_t pol=vm["pol"].as<size_t>();
    size_t j=0;
    for(size_t i=0; i<indata.nrow(); ++i)
    {
      if (indd(i)==cdd)
      {
	casa::Array<std::complex<float> > a;
	indata.get(i,a, casa::True);
	for (size_t k=0; k<nch; ++k)
	{
	  data[2*nch*j+2*k]  =a(casa::IPosition(2,pol,k)).real();
	  data[2*nch*j+2*k+1]=a(casa::IPosition(2,pol,k)).imag();
	}
	v_a1[j]=a1(i);
	v_a2[j]=a2(i);
	v_field[j]=field(i);
	v_time[j]=time(i);
	v_state[j]=state(i);
	++j;
      }
    }
    //Scale back the arrays to the correct size
    
    //! Need to the data array here
    v_a1.resize(j);
    v_a2.resize(j);
    v_field.resize(j);
    v_time.resize(j);
    v_state.resize(j);
    
    // No data to write for this SPW
    if (j==0)
      continue;

    hd5::write(fout,
	       group+"/VIS",
	       j, nch, 2,
	       data);
    hd5::write(fout,
	       group+"/ANTENNA1",
	       v_a1);
    hd5::write(fout,
	       group+"/ANTENNA2",
	       v_a2);
    hd5::write(fout,
	       group+"/FIELD",
	       v_field);
    hd5::write(fout,
	       group+"/TIME",
	       v_time);    
    hd5::write(fout,
	       group+"/STATE",
	       v_state);    
  }
  H5Gcreate(fout,
	    "/OBSCONF",
	    NULL, NULL, NULL);
  LibAIR::StateIntentMap si;
  LibAIR::scanIntents(ms, si);
  hd5::writeAtt(fout,
		std::string("/OBSCONF"),
		std::string("SCANINENT"),
		si);
}

void dumpWVR(const casa::MeasurementSet &ms,
	     hid_t fout,
	     const boost::program_options::variables_map &vm)
{
  using namespace LibAIR;

  std::set<size_t> WVRdd=WVRDataDescIDs(ms);

  casa::ROArrayColumn<casa::Complex> indata(ms, 
					    casa::MS::columnName(casa::MS::DATA));
  casa::ROScalarColumn<casa::Int> indd(ms,
				       casa::MS::columnName(casa::MS::DATA_DESC_ID));
  casa::ROScalarColumn<casa::Int> a1(ms,
				     casa::MS::columnName(casa::MS::ANTENNA1));
  casa::ROScalarColumn<casa::Int> a2(ms,
				     casa::MS::columnName(casa::MS::ANTENNA2));

  casa::ROScalarColumn<casa::Int> state(ms,
				     casa::MS::columnName(casa::MS::STATE_ID));

  casa::ROScalarColumn<casa::Double> time(ms,
					  casa::MS::columnName(casa::MS::TIME));  

  const size_t nrows=indata.nrow();

  AntSet wvrants=WVRAntennas(ms);

  for(AntSet::const_iterator ai= wvrants.begin();
      ai != wvrants.end();
      ++ai)
  {
    size_t A=*ai;
    // Vector to store the antenna values
    std::vector<int> v_a(nrows);
    // Time stamps
    std::vector<double> v_time(nrows);
    std::vector<int> v_state(nrows);
    
    // WVRs always have 4 channels
    const size_t nch=4;
    std::vector<double> data(nrows*nch);

    size_t j=0;
    for(size_t i=0; i<indata.nrow(); ++i)
    {
      if (a1(i)== A and 
	  a2(i)== A and
	  WVRdd.count(indd(i)) > 0)
      {
	v_a[j]=a1(i);
	v_time[j]=time(i);
	v_state[j]=state(i);

	casa::Array<std::complex<float> > a;
	indata.get(i,a, casa::True);
	for(size_t k=0; k<nch; ++k)
	{
	  data[nch*j+k]=a(casa::IPosition(2,k,0)).real();
	}
	++j;
      }
    }

    if (j==0)
    {
      std::cout<<"Tried to write out WVR data but didn't see any WVR data rows";
    }
    
    std::string agroup((boost::format("/ANT%i") % A).str());
    H5Gcreate(fout,
	      agroup.c_str(),
	      NULL, NULL, NULL);

    std::string group(agroup+"/WVR");
    H5Gcreate(fout,
	      group.c_str(),
	      NULL, NULL, NULL);

    // Scale back the array to right size
    v_time.resize(j);
    v_state.resize(j);
    hd5::write(fout,
	       group+"/SKY",
	       j,
	       nch,
	       data);
    hd5::write(fout,
	       group+"/TIME",
	       v_time);    
    hd5::write(fout,
	       group+"/STATE",
	       v_state);    

  }
}

void dumpAntennaTable(const casa::MeasurementSet &ms,
		      hid_t fout)
{
  using namespace LibAIR;
  antpos_t positions;
  getAntPos(ms, positions);
  aname_t  names=getAName(ms);

  H5Gcreate(fout,
	    "/Antennas",
	    NULL, NULL, NULL);

  hd5::write(fout,
	     "/Antennas/Positions",
	     positions.size1(),
	     positions.size2(),
	     &positions(0,0));
  
 } 

int main(int argc,  char* argv[])
{
  using namespace boost::program_options;

  options_description desc("Allowed options");
  desc.add_options()
    ("ms", value< std::vector<std::string> >(),
     "Input measurement set")
    ("output",
     value<std::string>(),
     "Name of the output file")
    ("spwconf",
     "Print information of the spectral window configurations")
    ("no-output",
     "Do not dump any data to hdf5, just info to stdout")
    ("infomain",
     "Print informartion about the main table")
    ("spw",
     value<size_t>(),
     "Spectral window to process")
    ("listwvrspws",
     "List information about WVR spectral windows")
    ("corrected",
     "Output corrected data column instead of just data")
    ("nowvr",
     "Do not output WVR data even if present (workaround for bugs option)")
    ("pol",
     value<size_t>()->default_value(0),
     "Output this polarisations number" )
    ("arraydata",
     "Output the contents of the array data structure only")
    ("help",
     "Print information about program usage")
    ;

  positional_options_description p;
  p.add("ms", -1);

  variables_map vm;
  store(command_line_parser(argc, argv).
	options(desc).positional(p).run(), 
 	vm);
  notify(vm);
  
  if (vm.count("help") or vm.count("ms")==0)
  {
    std::cout<<"Dump information from Measurement Set to HDF5"
      	     <<std::endl
	     <<std::endl
	     <<"GPL license -- you have the right to the source code. See COPYING"
	     <<std::endl
	     <<std::endl
	     <<desc;
    return 0;
  }

  casa:: MeasurementSet ms(vm["ms"].as<std::vector<std::string> >()[0]); 

  if (vm.count("spwconf"))
  {
    listSPWs(ms, std::cout);
  }
  if (vm.count("infomain"))
  {
    infoMain(ms, std::cout);
  }
  if (vm.count("listwvrspws"))
  {
    listWVRSPWs(ms, std::cout);
  }

  std::string fnameout;
  if (vm.count("output"))
  {
    fnameout=vm["output"].as<std::string>();
  }
  else
  {
    std::string msname=vm["ms"].as<std::vector<std::string> >()[0];
    boost::erase_tail(msname, 3);
    fnameout=msname+".hd5";
  }

  if (not vm.count("no-output"))
  {
    hid_t fout=H5Fcreate(fnameout.c_str(), 
			 H5F_ACC_TRUNC,
			 NULL, NULL);

    if (vm.count("arraydata"))
    {
      casa::MeasurementSet ms(vm["ms"].as<std::vector<std::string> >()[0]); 
      boost::scoped_ptr<LibAIR::InterpArrayData> d (LibAIR::loadWVRData(ms));
      H5Gcreate(fout,
		"/ArrayData",
		NULL, NULL, NULL);
      hd5::write(fout,
		 "/ArrayData/Time",
		 d->g_time());
      hd5::write(fout,
		 "/ArrayData/Elevation",
		 d->g_el());
      hd5::write(fout,
		 "/ArrayData/Azimuth",
		 d->g_az());
      hd5::write(fout,
		 "/ArrayData/WVR",
		 d->g_time().size(),
		 d->nWVRs,
		 4,
		 &(d->g_wvrdata()[0][0][0]));
    }
    else
    {
      dumpMain(ms,
	       fout,
	       vm);
      
      if (vm.count("nowvr")==0 and  LibAIR::nWVRSPWIDs(ms) > 0 )
      {
	// Only try to write WVR data if it is in the Measurement Set
      dumpWVR(ms,
	      fout,
	      vm);
      }
      
      
      dumpFQ(ms,
	     fout,
	     vm);
    }
    dumpAntennaTable(ms, fout);
    H5Fclose(fout);
  }
  return 0;
}
