/**
   Bojan Nikolic <b.nikolic@mrao.cam.ac.uk>, <bojan@bnikolic.co.uk>
   Initial version January 2010. 
   
   This file is part of LibAIR and is licensed under GNU Public
   License Version 2
   
   \file wvrgcal.cpp

   This is a command line program that reads WVR data from a CASA
   measurement set, computes the predicted complex gain of each
   antenna as a function of time from these WVR data and then writes
   these solutions out to a CASA gain table. 
   
*/

#include <iostream>

#include <boost/program_options.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/foreach.hpp>
#include <boost/tuple/tuple.hpp>

#include <ms/MeasurementSets/MeasurementSet.h>

#include "casawvr/mswvrdata.hpp"
#include "casawvr/msgaintable.hpp"
#include "casawvr/msutils.hpp"
#include "casawvr/msspec.hpp"
#include "casawvr/msantdata.hpp"
#include "src/apps/arraydata.hpp"
#include "src/apps/arraygains.hpp"
#include "src/apps/almaabs.hpp"
#include "src/apps/dtdlcoeffs.hpp"
#include "src/apps/almaresults.hpp"
#include "src/apps/segmentation.hpp"
#include "src/libair_main.hpp"

#ifdef BUILD_HD5
#include "../hd5wvr/resoutput.hpp"
#endif

#include "wvrgcalerrors.hpp"
#include "wvrgcalfeedback.hpp"

using LibAIR::fatalMsg;
using LibAIR::errorMsg;
using LibAIR::warnMsg;

/// Check the options and parameters supplied by the user for internal
/// consistency 
bool checkPars(const boost::program_options::variables_map &vm)
{
  if (vm.count("ms") <1)
  {
    fatalMsg("No input measurement sets given -- aborting ");
    return true;
  }

  if (vm.count("output") <1)
  {
    fatalMsg("No output file give -- aborting ");
    return true;
  }

  if (vm.count("segfield") )
  {
    fatalMsg("The --segfield option has been removed because it does not handle mosaic observations well.");
    fatalMsg("Please use the --segsource option instead");
    return true;
  }

  if (vm.count("segsource") && vm.count("cont"))
  {
    fatalMsg("Multiple retrievals using continuum estimation not yet supported");
    fatalMsg("Use only one of  --sergfield OR --cont");
    return true;
  }

  if (vm["nsol"].as<int>()>1 && vm.count("cont"))
  {
    fatalMsg("Multiple retrievals using continuum estimation not yet supported");
    fatalMsg("You can not use the nsol parameter yet with cont");
    return true;
  }

  if (vm.count("reverse") and vm.count("reversespw"))
  {
    warnMsg("You are specifying both the reverse and reversespw options;"
	    "the latter will be ignored and all spectral windows will be reversed");
  }
  return false;

  if (vm.count("smooth") and vm["smooth"].as<int>() < 1)
  {
    warnMsg("Smooth parameter must be 1 or greater");
    return true;
  }
  return false;
  
}

void checkWarnPars(const boost::program_options::variables_map &vm)
{

  if (vm.count("statfield"))
  {
    warnMsg("The use of \"statfield\" is not recommended as mosaiced"
	    "observations are recorded as many separate fields. Recomended option"
	    "to use is \"statsource\". ");
  }
}

/* Checks on parameter that can only be done once the measurement set
   is opened.  Do not put computatntially intensive checks otherwise
   feedback to the user will be too slow.
*/
void checkMSandPars(const casa::MeasurementSet &ms,
		    const boost::program_options::variables_map &vm)
{
  if (vm.count("statsource"))
  {
    std::string srcname=vm["statsource"].as<std::vector<std::string> >()[0];
    std::set<size_t> fselect=LibAIR::getSrcFields(ms,
						  srcname);
    if (fselect.size() == 0)
    {
      std::cout<<"WARNING: No fields appear to be identified with source "<<srcname <<std::endl
	       <<"         that you supplied to the statfield option"<<std::endl
	       <<"         Statistics will be corrupted and wvrgcal may fail"<<std::endl
	       <<std::endl;
    }
  }
}

/** \brief Take a parameter than can be specified as a sequence of
    antenna numbers or names and always return as a sequence of
    antenna numbers.
 */
std::set<int> getAntPars(const std::string &s,
			 const boost::program_options::variables_map &vm,
			 const casa::MeasurementSet &ms)
{
  using namespace LibAIR;
  aname_t anames=getAName(ms);
  std::vector<std::string> pars=vm[s].as<std::vector<std::string> >();
  std::set<int> res;
  for (size_t i=0; i<pars.size(); ++i)
  {
    if (anames.right.count(pars[i]))
    {
      res.insert(anames.right.at(pars[i]));
    }
    else
    {
      // should be an antenna number
      try {
	int n=boost::lexical_cast<int>(pars[i]);
	if (anames.left.count(n) ==0)
	{
	  throw AntIDError(n,
			   anames);
	}
	res.insert(n);
      }
      catch (const boost::bad_lexical_cast & bc)
      {
	throw AntIDError(pars[i],
			 anames);

      }

    }
  }
  return res;
}


/** Simple function to turn the command line parameters into a single
    string.

    Return by value as only needs to be called once in program
*/
static std::string buildCmdLine(int argc,
				char* argv[])
{
  std::string cmdline;
  for (size_t i=0; i< (size_t)argc; ++i)
  {
    cmdline+=std::string(argv[i]);
    cmdline+=" ";
  }
  return cmdline;
}

/** \brief Flag and interpolate WVR data
 */
void flagInterp(const casa::MeasurementSet &ms,
		const std::set<int> &wvrflag,
		LibAIR::InterpArrayData &d)
{
  LibAIR::antpos_t apos;
  LibAIR::getAntPos(ms, apos);
  LibAIR::AntSet wvrflag_s(wvrflag.begin(), 
			   wvrflag.end());
  for(std::set<int>::const_iterator i=wvrflag.begin();
      i!=wvrflag.end(); 
      ++i)
  {
    LibAIR::AntSetWeight near=LibAIR::linNearestAnt(apos, 
						    *i, 
						    wvrflag_s, 
						    3);
    LibAIR::interpBadAntW(d, 
			  *i, 
			  near);
  }
  
}

/// Work out which spectral windows might need to be reversed
std::set<size_t> reversedSPWs(const LibAIR::MSSpec &sp,
			      const boost::program_options::variables_map &vm)
{
  std::set<size_t> reverse;
  if (vm.count("reverse"))
  {
    for (size_t spw =0; spw<sp.spws.size(); ++spw)
      reverse.insert(spw);
  }    
  if (vm.count("reversespw"))
  {
    std::vector<int> torev=vm["reversespw"].as<std::vector<int> >();
    for(size_t i=0; i<torev.size(); ++i)
    {
      if (torev[i]<0 or torev[i] >= (int)sp.spws.size())
      {
	throw LibAIR::SPWIDError(torev[i], 
				 sp.spws.size());
      }
      reverse.insert(torev[i]);
    }
  }    
  return reverse;
}

void printExpectedPerf(const LibAIR::ArrayGains &g,
		       const LibAIR::dTdLCoeffsBase &coeffs,
		       const std::vector<std::pair<double, double> > &tmask)
{

  std::cout<<"  Expected performance "<<std::endl
	   <<"------------------------------------------------------------------"<<std::endl;
  
  std::vector<double> cr, err;
  coeffs.repr(cr, err);
  std::cout<<"* Estimated WVR thermal contribution to path fluctuations (micron per antenna): "
	   <<LibAIR::thermal_error(cr)/1e-6
	   <<std::endl;
  const double grmsbl=g.greatestRMSBl(tmask);
  std::cout<<"* Greatest Estimated path fluctuation is (micron on a baseline): "
	   <<grmsbl/1e-6
	   <<std::endl;
  std::cout<<"* Rough estimate path error due to coefficient error (micron on a baseline): "
	   <<grmsbl* (err[1]/cr[1])/1e-6
	   <<std::endl
    	   <<std::endl;
}

/** Compute the time intervals over which the statistics should be
    computed
 */
void statTimeMask(const casa::MeasurementSet &ms,
		  const boost::program_options::variables_map &vm,
		  std::vector<std::pair<double, double> > &tmask)
{
  std::vector<int> flds;
  std::vector<double> time;
  std::vector<int> src;
  LibAIR::fieldIDs(ms, 
		   time,
		   flds,
		   src);
  std::vector<size_t> spws;
  LibAIR::dataSPWs(ms, spws);

  if (vm.count("statfield") == 0 && vm.count("statsource") == 0)
  {
    tmask.resize(0);
    tmask.push_back(std::pair<double, double>(time[0], time[time.size()-1]));
  }
  else if ( vm.count("statsource") > 0)
  {
    std::set<size_t> fselect=LibAIR::getSrcFields(ms,
						  vm["statsource"].as<std::vector<std::string> >()[0]);
    LibAIR::fieldTimes(time,
		       flds,
		       spws,
		       fselect,
		       0,
		       tmask);    
  }
  else
  {
    std::vector<std::string> fields=vm["statfield"].as<std::vector<std::string> >();
    LibAIR::field_t fnames=LibAIR::getFieldNames(ms);

    std::set<size_t> fselect;
    if (fnames.right.count(fields[0])) // User supplied  field *name*
    {
      size_t n=fnames.right.at(fields[0]);
      fselect.insert(n);
    }
    else
    {
      try 
      {
	size_t n=boost::lexical_cast<int>(fields[0]);
	fselect.insert(n);
      }
      catch (const boost::bad_lexical_cast & bc)
      {
	std::cout<<"Warning: Could not understand statfield argument. Will use zeroth field."
		 <<std::endl;
      }
    }
    LibAIR::fieldTimes(time,
		       flds,
		       spws,
		       fselect,
		       0,
		       tmask);
  }
  LibAIR::printStatTimes(std::cout,
			 time,
			 tmask);
}
		  

/// Compute the discrepance in path estimate between channels 1 and 3
void computePathDisc(const LibAIR::InterpArrayData &d,
		     const std::vector<std::pair<double, double> > &tmask,
		     LibAIR::dTdLCoeffsBase  &coeffs,
		     std::vector<double> &res)
{
  LibAIR::ArrayGains g1(d.g_time(), 
			d.g_el(),
			d.g_state(),
			d.g_field(),
			d.g_source(),
			d.nWVRs);
  boost::array<double, 4> c1mask = {{0, 1, 0,0}};
  boost::array<double, 4> c3mask = {{0, 0, 0,1}};
  boost::array<double, 4> callmask = {{1, 1, 1,1}};
    
  coeffs.chmask=c1mask;
  g1.calc(d,
	  coeffs);    
  
  LibAIR::ArrayGains g3(d.g_time(), 
			d.g_el(),
			d.g_state(),
			d.g_field(),
			d.g_source(),
			d.nWVRs);
  coeffs.chmask=c3mask;
  g3.calc(d,
	  coeffs);    

  g1.pathDiscAnt(g3, 
		 tmask,
		 res);

  coeffs.chmask=callmask;
  
}

void printFieldSegments(const std::vector<std::pair<double, double> >  &fb,
			double tbase)
{
  for (size_t i=0; i<fb.size(); ++i)
    {
      std::cout<<fb[i].first-tbase<<","<<fb[i].second-tbase<<std::endl;
    }

}

std::vector<std::set<std::string> > getTied(const boost::program_options::variables_map &vm)
{
  using namespace boost::algorithm;
  std::vector<std::set<std::string> > res;
  for (size_t i=0; i<vm.count("tie"); ++i)
  {
    std::set<std::string> cs;
    std::string par(vm["tie"].as<std::vector<std::string> >()[i]);
    split(cs, par, is_any_of(","));
    res.push_back(cs);
  }
  return res;
}

// Convert tied source names to tied source IDs
std::vector<std::set<size_t> >  tiedIDs(const std::vector<std::set<std::string> > &tied,
					const casa::MeasurementSet &ms)
{
  boost::bimap<size_t, std::string > srcmap=LibAIR::getSourceNames(ms);
  std::vector<std::set<size_t> > res;
  for (size_t i=0; i<tied.size(); ++i)
  {
    std::set<size_t> cs;
    for(std::set<std::string>::const_iterator j=tied[i].begin();
	j!=tied[i].end();
	++j)
      {	
	try
	  {
	    int srcid=boost::lexical_cast<int>(*j);
	    cs.insert(srcid);
	  }
	catch (const boost::bad_lexical_cast & bc)
	  {
	    cs.insert(srcmap.right.at(*j));
	  }
      }
    res.push_back(cs);
  }
  return res;
}

void printTied(const std::vector<std::set<std::string> > &tied,
	       const std::vector<std::set<size_t> > &tiedi)
{
  for(size_t i=0; i<tied.size(); ++i)
  {
    std::cout<<"Tying: ";
    BOOST_FOREACH(const std::string &x, tied[i])
      std::cout<<x<<" and ";
    std::cout<<std::endl;
  }
  if (tied.size())
    std::cout<<"Tied sets as numerical source IDs:"<<std::endl;
  for(size_t i=0; i<tiedi.size(); ++i)
  {
    std::cout<<"Tying: ";
    BOOST_FOREACH(const size_t &x, tiedi[i])
      std::cout<<x<<" and ";
    std::cout<<std::endl;
  }

}

/** Compute the set of source_ids corresponding to a vector of source
    names.
 */
std::set<size_t> sourceSet(const std::vector<std::string> &sources,
			   const casa::MeasurementSet &ms)
{
  boost::bimap<size_t, std::string > snames=LibAIR::getSourceNames(ms);
  std::set<size_t> sset;
  for(size_t i=0; i<sources.size(); ++i)
    sset.insert(snames.right.at(sources[i]));  
  return sset;
}
  

/** Filter the set of input WVR measurements to retrieve the
    coefficients from to exclude flagged sources

    This function takes and returns two containers: the first is the
    list of WVR values to be analysed; the second is a vector of time
    ranges to use.
    
    These time ranges must be filtered together with the inputs since
    they are not referenced directly to the inputs are simply assumed
    to be "row-synchronous".
    
 */
std::pair<LibAIR::ALMAAbsInpL,  std::vector<std::pair<double, double> > >
filterInp(const LibAIR::ALMAAbsInpL &inp,
	  const std::vector<std::pair<double, double> > &fb,
	  const std::vector<std::string> &sourceflag,
	  const casa::MeasurementSet &ms)
{

  std::set<size_t> flagset=sourceSet(sourceflag, ms);

  LibAIR::ALMAAbsInpL res;
  std::vector<std::pair<double, double> > rfb;
  size_t j=0;
  for(LibAIR::ALMAAbsInpL::const_iterator i=inp.begin();
      i!=inp.end();
      ++i, ++j)
  {
    if (flagset.count(i->source) ==0)
    {
      res.push_back(*i);
      rfb.push_back(fb[j]);
    }
  }
  return std::make_pair(res, rfb);
}

/** Return the set of antenna IDs that do not have a WVR
 */
std::set<int> NoWVRAnts(const LibAIR::aname_t &an)
{
  std::set<int> res;
  for(LibAIR::aname_t::const_iterator i=an.begin();
      i!= an.end();
      ++i)
  {
    if (i->right[0]=='C' and i->right[1]=='M')
      res.insert(i->left);
  }
  return res;
}


static void defineOptions(boost::program_options::options_description &desc,
			  boost::program_options::positional_options_description &p)
{
  using namespace boost::program_options;
  desc.add_options()
    ("ms", value< std::vector<std::string> >(),
     "Input measurement set")
    ("output",
     value<std::string>(),
     "Name of the output file")
    ("toffset",
     value<double>()->default_value(0),
     "Time offset (in seconds) between interferometric and WVR data")
    ("nsol",
     value<int>()->default_value(1),
     "Number of solutions for phase correction coefficients to make during this observation")
    ("help",
     "Print information about program usage")
    ("segfield",
     "Do a new coefficient calculation for each field (this option is disabled, see segsource)")
    ("segsource",
     "Do a new coefficient calculation for each source")
    ("reverse",
     "Reverse the sign of correction in all SPW (e.g. due to AIV-1740)")
    ("reversespw",
     value< std::vector<int> >(),
     "Reverse the sign correction for this spw")
    ("disperse",
     "Apply correction for dispersion")
    ("cont",
     "Estimate the continuum (e.g., due to clouds)")
    ("wvrflag",
     value< std::vector<std::string> >(),
     "Flag this WVR (labelled with either antenna number or antenna name) as bad, and replace its data with interpolated values ")
    ("sourceflag",
     value< std::vector<std::string> >(),
     "Flag the WVR data for this source and do not produce any phase corrections on it")
    ("statfield",
     value< std::vector<std::string> >(),
     "Compute the statistics (Phase RMS, Disc) on this field only"
     )
    ("statsource",
     value< std::vector<std::string> >(),
     "Compute the statistics (Phase RMS, Disc) on this source only"
     )
    ("tie",
     value< std::vector<std::string> >(),
     "Prioritise tieing the phase of these sources as well as possible")
    ("smooth",
     value<int>(),
     "Smooth WVR data by this many samples before applying the correction")
    ("scale",
     value<double>()->default_value(1.0),
     "Scale the entire phase correction by this factor")
    ;
  p.add("ms", -1);
}

int main(int argc,  char* argv[])
{
  using namespace boost::program_options;

  options_description desc("Allowed options");
  positional_options_description p;
  defineOptions(desc, p);


  variables_map vm;
  store(command_line_parser(argc, argv).
	options(desc).positional(p).run(), 
 	vm);
  notify(vm);

  LibAIR::printBanner(std::cout);


  
  if (vm.count("help"))
  {
    std::cout<<"Write out a gain table based on WVR data"
      	     <<std::endl
	     <<std::endl
	     <<"GPL license -- you have the right to the source code. See COPYING"
	     <<std::endl
	     <<std::endl
	     <<desc;
    return 0;
  }

  if (checkPars(vm))
  {
    return -1;
  }

  checkWarnPars(vm);
  std::vector<std::set<std::string> > tied=getTied(vm);

  std::string msname(vm["ms"].as<std::vector<std::string> >()[0]);
  casa::MeasurementSet ms(msname); 

  checkMSandPars(ms, vm);

  std::string fnameout=vm["output"].as<std::string>();

  std::set<size_t> useID=LibAIR::skyStateIDs(ms);
  boost::scoped_ptr<LibAIR::InterpArrayData> d (LibAIR::loadWVRData(ms));
  d->offsetTime(vm["toffset"].as<double>());

  if (vm.count("smooth"))
  {
    smoothWVR(*d, vm["smooth"].as<int>());
  }
  
  d.reset(LibAIR::filterState(*d, useID));

  std::set<int> wvrflag;
  // Flag and interpolate
  if (vm.count("wvrflag"))
  {
    wvrflag=getAntPars("wvrflag", vm, ms);    
  }
  LibAIR::aname_t anames=LibAIR::getAName(ms);
  std::set<int> nowvr=NoWVRAnts(anames);

  std::set<int> interpwvrs(wvrflag);
  interpwvrs.insert(nowvr.begin(), nowvr.end());
  flagInterp(ms,
             interpwvrs,
             *d);



  LibAIR::ArrayGains g(d->g_time(), 
		       d->g_el(),
		       d->g_state(),
		       d->g_field(),
		       d->g_source(),
		       d->nWVRs);

  boost::scoped_ptr<LibAIR::dTdLCoeffsBase>  coeffs;

  // These are the segments on which coefficients are re-calculated
  std::vector<std::pair<double, double> >  fb;

  if ( vm.count("cont") )
  {
    std::cout<<"[Output from \"cont\" option has not yet been updated]"
	     <<std::endl;
    coeffs.reset(LibAIR::SimpleSingleCont(*d));
  }
  else
  {
    LibAIR::ALMAAbsInpL inp;
    if (vm.count("segsource"))
    {
      std::vector<int> flds;
      std::vector<double> time;
      std::vector<int> src;
      LibAIR::fieldIDs(ms, 
		       time,
		       flds,
		       src);
      std::vector<std::set<size_t> >  tiedi=tiedIDs(tied, ms);
      printTied(tied, tiedi);
      LibAIR::fieldSegmentsTied(time,
				src,
				tiedi,
				fb);
      //printFieldSegments(fb, time[0]);
      inp=FieldMidPointI(*d,
			 time,
			 fb,
			 useID);
    }
    else
    {
      const size_t n=vm["nsol"].as<int>();
      inp=LibAIR::MultipleUniformI(*d, 
				   n,
				   useID);
    }

    if (vm.count("sourceflag"))
    {
      boost::tie(inp,fb)=filterInp(inp,
				   fb,
				   vm["sourceflag"].as<std::vector<std::string> >(),
				   ms);
    }

    std::cout<<inp<<std::endl;

    std::cerr<<"Calculating the coefficients now...";
    boost::ptr_list<LibAIR::ALMAResBase> rlist=LibAIR::doALMAAbsRet(inp);
    std::cerr<<"done!"
	     <<std::endl;

    std::cout<<"       Retrieved parameters      "<<std::endl
	     <<"----------------------------------------------------------------"<<std::endl
	     <<rlist<<std::endl;

    if (vm.count("segsource"))
    {
      std::vector<int> flds;
      std::vector<double> time;
      std::vector<int> src;
      LibAIR::fieldIDs(ms, 
		       time,
		       flds,
		       src);
      coeffs.reset(LibAIR::SimpleMultiple(*d, 
					  time,
					  fb,
					  rlist));   
    }
    else
    {
      coeffs.reset(LibAIR::ALMAAbsProcessor(inp, rlist));
    }
  }

  if (coeffs->isnan())
  {
    LibAIR::printNoSolution(std::cerr);
    return -1;
  }
  g.calc(*d,
	 *coeffs);    

  if (vm.count("sourceflag"))
  {
    std::set<size_t> flagset=sourceSet(vm["sourceflag"].as<std::vector<std::string> >(),
				       ms);
    g.blankSources(flagset);
  }

  

  std::vector<std::pair<double, double> > tmask;
  statTimeMask(ms, vm, tmask);

  std::vector<double> pathRMS;
  g.pathRMSAnt(tmask, pathRMS);

  
  
  std::vector<double> pathDisc;
  computePathDisc(*d, 
		  tmask,
		  *coeffs,
		  pathDisc);
  
  std::cout<<LibAIR::AntITable(anames,
			       wvrflag,
                               nowvr,
			       pathRMS,
			       pathDisc);

  printExpectedPerf(g, 
		    *coeffs,
		    tmask);

  if (vm.count("scale"))
  {
    g.scale(vm["scale"].as<double>());
  }
    
  LibAIR::MSSpec sp;
  loadSpec(ms, sp);
  std::set<size_t> reverse=reversedSPWs(sp, vm);  

  LibAIR::writeGainTbl(g,
		       fnameout.c_str(),
		       sp,
		       reverse,
		       vm.count("disperse")>0,
		       msname
		       );

#ifdef BUILD_HD5
  LibAIR::writeAntPath(g,
		       fnameout+".hd5");
#endif

  LibAIR::addCalHistory(fnameout.c_str(),
			buildCmdLine(argc,
				     argv));

  return 0;
}
